from .backend import Backend, CachedBackend, InMemoryCachedBackend

try:
    from ._dynamodb import DynamoDBBackend
except ImportError:
    pass
from ._networkx import NetworkXBackend
from ._dataframe import DataFrameBackend

try:
    from ._sqlbackend import SQLBackend
except ImportError:
    pass

try:
    from ._networkit import NetworkitBackend
except ImportError:
    pass

__all__ = [
    "Backend",
    "CachedBackend",
    "InMemoryCachedBackend",
    "NetworkXBackend",
    "DataFrameBackend",
    "DynamoDBBackend",
    "SQLBackend",
    "NetworkitBackend",
]



import time
from .backend import InMemoryCachedBackend
from ._networkx import NetworkXBackend
from ._sqlbackend import SQLBackend


def test_can_create_cached_backend():
    InMemoryCachedBackend(NetworkXBackend(), maxsize=1024, ttl=20)


def test_can_add_node():
    vanilla = NetworkXBackend()
    cached = InMemoryCachedBackend(vanilla, maxsize=1024, ttl=20)
    assert vanilla.get_node_count() == 0
    assert cached.get_node_count() == 0
    cached.add_node("a", {})
    assert vanilla.get_node_count() == 1
    assert cached.get_node_count() == 1


def test_added_node_ignored_when_no_dirty_on_write():
    vanilla = NetworkXBackend()
    cached = InMemoryCachedBackend(
        vanilla, dirty_cache_on_write=False, maxsize=1024, ttl=20
    )
    assert vanilla.get_node_count() == 0
    assert cached.get_node_count() == 0
    cached.add_node("a", {})
    assert vanilla.get_node_count() == 1
    assert cached.get_node_count() == 0


def test_can_add_nodes():
    cached = InMemoryCachedBackend(NetworkXBackend(), maxsize=1024, ttl=20)
    assert cached.get_node_count() == 0
    cached.add_node("a", {})
    assert cached.get_node_count() == 1
    cached.add_node("a", {})
    assert cached.get_node_count() == 1
    cached.add_node("b", {})
    assert cached.get_node_count() == 2


def test_cache_is_faster_than_no_cache():
    cached = InMemoryCachedBackend(SQLBackend(), maxsize=1024, ttl=20)
    for i in range(1000):
        cached.add_node(i, {})

    tic = time.time()
    node_count = cached.get_node_count()
    toc = time.time()

    tic2 = time.time()
    node_count2 = cached.get_node_count()
    toc2 = time.time()

    # Dirty the cache:
    cached.add_node(1000, {})

    tic3 = time.time()
    node_count3 = cached.get_node_count()
    toc3 = time.time()

    assert node_count == node_count2 == (node_count3 - 1)
    assert (toc - tic) > (toc2 - tic2)
    assert (toc3 - tic3) > (toc2 - tic2)


def test_cache_info():
    cached = InMemoryCachedBackend(SQLBackend(), maxsize=1024, ttl=20)
    cached.add_node("foo", {})
    assert cached.cache_info()["get_node_count"].misses == 0
    assert cached.cache_info()["get_node_count"].hits == 0

    cached.get_node_count()

    assert cached.cache_info()["get_node_count"].misses == 1
    assert cached.cache_info()["get_node_count"].hits == 0

    cached.get_node_count()

    assert cached.cache_info()["get_node_count"].misses == 1
    assert cached.cache_info()["get_node_count"].hits == 1



import unittest

from . import Graph, DiGraph


class TestGraph(unittest.TestCase):
    def test_can_create(self):
        Graph()

    def test_can_use_nx_backend(self):
        Graph().nx

    def test_can_create_directed(self):
        assert Graph(directed=True).nx.is_directed() is True
        assert Graph(directed=False).nx.is_directed() is False
        assert DiGraph().nx.is_directed() is True



import cachetools.func
from typing import Callable, Hashable, Collection
import abc

import pandas as pd


class Backend(abc.ABC):
    """
    Abstract base class for the management of persisted graph structure.

    Do not use this class directly.

    """

    def __init__(self, directed: bool = False):
        """
        Create a new Backend instance.

        Arguments:
            directed (bool: False): Whether to make the backend graph directed

        Returns:
            None

        """
        ...

    def ingest_from_edgelist_dataframe(
        self, edgelist: pd.DataFrame, source_column: str, target_column: str
    ) -> None:
        """
        Ingest an edgelist from a Pandas DataFrame.

        """
        ...

    def is_directed(self) -> bool:
        """
        Return True if the backend graph is directed.

        Arguments:
            None

        Returns:
            bool: True if the backend graph is directed.

        """
        ...

    def add_node(self, node_name: Hashable, metadata: dict):
        """
        Add a new node to the graph.

        Arguments:
            node_name (Hashable): The ID of the node
            metadata (dict: None): An optional dictionary of metadata
            upsert (bool: True): Update the node if it already exists. If this
                is set to False and the node already exists, a backend may
                choose to throw an error or proceed gracefully.

        Returns:
            Hashable: The ID of this node, as inserted

        """
        ...

    def add_nodes_from(self, nodes_for_adding, **attr):
        """
        Add nodes to the graph.

        Arguments:
            nodes_for_adding: nodes to add
            attr: additional attributes
        """
        for node, metadata in nodes_for_adding:
            self.add_node(node, {**attr, **metadata})

    def get_node_by_id(self, node_name: Hashable):
        """
        Return the data associated with a node.

        Arguments:
            node_name (Hashable): The node ID to look up

        Returns:
            dict: The metadata associated with this node

        """
        ...

    def all_nodes_as_iterable(self, include_metadata: bool = False) -> Collection:
        """
        Get a generator of all of the nodes in this graph.

        Arguments:
            include_metadata (bool: False): Whether to include node metadata in
                the response

        Returns:
            Generator: A generator of all nodes (arbitrary sort)

        """
        ...

    def has_node(self, u: Hashable) -> bool:
        """
        Return true if the node exists in the graph.

        Arguments:
            u (Hashable): The ID of the node to check

        Returns:
            bool: True if the node exists
        """
        try:
            return self.get_node_by_id(u) is not None
        except KeyError:
            return False

    def has_edge(self, u: Hashable, v: Hashable) -> bool:
        """
        Return true if the edge exists in the graph.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID

        Returns:
            bool: True if the edge exists
        """
        try:
            return self.get_edge_by_id(u, v) is not None
        except KeyError:
            return False

    def add_edge(self, u: Hashable, v: Hashable, metadata: dict):
        """
        Add a new edge to the graph between two nodes.

        If the graph is directed, this edge will start (source) at the `u` node
        and end (target) at the `v` node.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID
            metadata (dict): Optional metadata to associate with the edge

        Returns:
            Hashable: The edge ID, as inserted.

        """
        ...

    def add_edges_from(self, ebunch_to_add, **attr):
        """
        Add new edges to the graph.

        Arguments:
            ebunch_to_add: list of (source, target, metadata)
            attr: additional common attributes
        """
        for u, v, metadata in ebunch_to_add:
            self.add_edge(u, v, {**attr, **metadata})

    def all_edges_as_iterable(self, include_metadata: bool = False) -> Collection:
        """
        Get a list of all edges in this graph, arbitrary sort.

        Arguments:
            include_metadata (bool: False): Whether to include edge metadata

        Returns:
            Generator: A generator of all edges (arbitrary sort)

        """
        ...

    def get_edge_by_id(self, u: Hashable, v: Hashable):
        """
        Get an edge by its source and target IDs.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID

        Returns:
            dict: Metadata associated with this edge

        """
        ...

    def get_node_successors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Collection:
        return self.get_node_neighbors(u, include_metadata)

    def get_node_neighbors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Collection:
        """
        Get a generator of all downstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        ...

    def get_node_predecessors(
        self, u: Hashable, include_metadata: bool = False
    ) -> Collection:
        """
        Get a generator of all upstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """
        ...

    def get_node_count(self) -> int:
        """
        Get an integer count of the number of nodes in this graph.

        Arguments:
            None

        Returns:
            int: The count of nodes

        """
        return len([i for i in self.all_nodes_as_iterable()])

    def get_edge_count(self) -> int:
        """
        Get an integer count of the number of edges in this graph.

        Arguments:
            None

        Returns:
            int: The count of edges

        """
        return len([i for i in self.all_edges_as_iterable()])

    def degree(self, u: Hashable) -> int:
        """
        Get the degree of a node.

        Arguments:
            u (Hashable): The node ID

        Returns:
            int: The degree of the node

        """
        return len([i for i in self.get_node_neighbors(u)])

    def degrees(self, nbunch=None) -> Collection:
        return {
            node: self.degree(node) for node in (nbunch or self.all_nodes_as_iterable())
        }

    def in_degree(self, u: Hashable) -> int:
        """
        Get the in-degree of a node.

        Arguments:
            u (Hashable): The node ID

        Returns:
            int: The in-degree of the node

        """
        return len(list(self.get_node_predecessors(u)))

    def in_degrees(self, nbunch=None) -> Collection:
        nbunch = nbunch or self.all_nodes_as_iterable()
        if isinstance(nbunch, (list, tuple)):
            return {node: self.in_degree(node) for node in nbunch}
        else:
            return self.in_degree(nbunch)

    def out_degree(self, u: Hashable) -> int:
        """
        Get the out-degree of a node.

        Arguments:
            u (Hashable): The node ID

        Returns:
            int: The out-degree of the node

        """
        return len(list(self.get_node_successors(u)))

    def out_degrees(self, nbunch=None) -> Collection:
        nbunch = nbunch or self.all_nodes_as_iterable()
        if isinstance(nbunch, (list, tuple)):
            return {node: self.out_degree(node) for node in nbunch}
        else:
            return self.out_degree(nbunch)


class CachedBackend(Backend):
    """
    A proxy Backend that serves as a cache for any other grand.Backend.

    """

    def __init__(self, backend: Backend): ...


class InMemoryCachedBackend(CachedBackend):
    """
    A proxy Backend that serves as a cache for any other grand.Backend.

    Wraps each call to the Backend with an LRU cache.

    """

    _cache_types = {
        "LRUCache": cachetools.func.lru_cache,
        "TTLCache": cachetools.func.ttl_cache,
        "LFUCache": cachetools.func.lfu_cache,
    }

    _default_uncacheable_methods = [
        "add_node",
        "add_nodes_from",
        "add_edge",
        "add_edges_from",
        "ingest_from_edgelist_dataframe",
    ]

    _default_write_methods = [
        "add_node",
        "add_nodes_from",
        "add_edge",
        "add_edges_from",
        "ingest_from_edgelist_dataframe",
    ]

    def __init__(
        self,
        backend: Backend,
        dirty_cache_on_write: bool = True,
        cache_type: str = "TTLCache",
        uncacheable_methods: list = None,
        write_methods: list = None,
        **cache_kwargs,
    ):
        """
        Initialize a new in-memory cache, using the cachetools library.

        Arguments:
            backend (grand.Backend): The backend to cache
            dirty_cache_on_write (bool): Whether to clear the cache on writes
            cache_type (str: "TTLCache"): The cache type to use. One of
                ["LRUCache", "TTLCache"]
            **cache_kwargs: Additional arguments to pass to the cache


        """
        self.backend = backend
        self._dirty_cache_on_write = dirty_cache_on_write
        self._uncacheable_methods = (
            uncacheable_methods or self._default_uncacheable_methods
        )
        self._write_methods = write_methods or self._default_write_methods
        if cache_type not in self._cache_types:
            raise ValueError(
                f"Unknown cache type: {cache_type}. "
                f"Valid types are: {self._cache_types.keys()}"
            )
        self._cache_factory = lambda: self._cache_types[cache_type](**cache_kwargs)

        self._method_lookup = {}

        def _dirty_cache_decorator(method_: Callable):
            def dirty_cache_dec_wrapper(*args, **kwargs):
                self.clear_cache()
                return method_(*args, **kwargs)

            return dirty_cache_dec_wrapper

        method_list = [
            attribute
            for attribute in dir(self.backend)
            if callable(getattr(self.backend, attribute))
            and not attribute.startswith("_")
        ]
        for method in method_list:
            if method in self._uncacheable_methods:
                setattr(self, method, getattr(self.backend, method))
            else:
                wrapped = self._wrapped(method)
                self._method_lookup[method] = wrapped
                setattr(self, method, wrapped)

            if self._dirty_cache_on_write and method in self._write_methods:
                setattr(
                    self, method, _dirty_cache_decorator(getattr(self.backend, method))
                )

    def _wrapped(self, method: str) -> Callable:
        c = self._cache_factory()(getattr(self.backend, method))
        return c

    def clear_cache(self):
        """
        Clear the cache.

        """
        for _, method in self._method_lookup.items():
            method.cache_clear()

    def cache_info(self):
        return {
            method_name: method.cache_info()
            for method_name, method in self._method_lookup.items()
        }



import pytest
from .metadatastore import DictMetadataStore, NodeNameManager


def test_can_create_dict_metadata_store():
    store = DictMetadataStore()
    assert store is not None


def test_can_add_to_dict_metadata_store():
    store = DictMetadataStore()
    store.add_node("a", {"a": 1})
    assert store.get_node("a") == {"a": 1}


def test_can_add_edge():
    store = DictMetadataStore()
    store.add_edge("a", "b", {"b": 2})
    assert store.get_edge("a", "b") == {"b": 2}


def test_can_update_edge():
    store = DictMetadataStore()
    store.add_edge("a", "b", {"b": 2})
    store.add_edge("a", "b", {"x": "x"})
    store.add_edge("a", "b", {"z": "z"})
    assert store.get_edge("a", "b")["b"] == 2
    assert store.get_edge("a", "b")["x"] == "x"
    assert store.get_edge("a", "b")["z"] == "z"


def test_cannot_get_invalid_node():
    store = DictMetadataStore()
    with pytest.raises(KeyError):
        store.get_node("a")


def test_can_create_dict_name_manager():
    store = NodeNameManager()
    assert store is not None


def test_can_add_name_manager():
    store = NodeNameManager()
    store.add_node("a", 1)
    assert "a" in store
    assert 1 not in store
    assert store.get_name(1) == "a"


def test_can_reverse_lookup_node_name_manager():
    store = NodeNameManager()
    store.add_node("a", 1)
    assert "a" in store
    assert 1 not in store
    assert store.get_id("a") == 1



import pytest
import os

import networkx as nx

from . import NetworkXBackend, DataFrameBackend

try:
    from ._dynamodb import DynamoDBBackend

    _CAN_IMPORT_DYNAMODB = True
except ImportError:
    _CAN_IMPORT_DYNAMODB = False

try:
    from ._igraph import IGraphBackend

    _CAN_IMPORT_IGRAPH = True
except ImportError:
    _CAN_IMPORT_IGRAPH = False

try:
    from ._networkit import NetworkitBackend

    _CAN_IMPORT_NETWORKIT = True
except ImportError:
    _CAN_IMPORT_NETWORKIT = False

try:
    from ._sqlbackend import SQLBackend

    _CAN_IMPORT_SQL = True
except ImportError:
    _CAN_IMPORT_SQL = False

from .. import Graph


backend_test_params = [
    pytest.param(
        (NetworkXBackend, {}),
        marks=pytest.mark.skipif(
            os.environ.get("TEST_NETWORKXBACKEND", default="1") != "1",
            reason="NetworkX Backend skipped because $TEST_NETWORKXBACKEND != 1.",
        ),
        id="NetworkXBackend",
    ),
]
backend_test_params = [
    pytest.param(
        (DataFrameBackend, {}),
        marks=pytest.mark.skipif(
            os.environ.get("TEST_DATAFRAMEBACKEND", default="1") != "1",
            reason="DataFrameBackend skipped because $TEST_DATAFRAMEBACKEND != 1.",
        ),
        id="DataFrameBackend",
    ),
]

if _CAN_IMPORT_DYNAMODB:
    backend_test_params.append(
        pytest.param(
            (DynamoDBBackend, {}),
            marks=pytest.mark.skipif(
                os.environ.get("TEST_DYNAMODB", default="1") != "1",
                reason="DynamoDB Backend skipped because $TEST_DYNAMODB != 0 or boto3 is not installed",
            ),
            id="DynamoDBBackend",
        ),
    )

if _CAN_IMPORT_SQL:
    backend_test_params.append(
        pytest.param(
            (SQLBackend, {"db_url": "sqlite:///:memory:"}),
            marks=pytest.mark.skipif(
                os.environ.get("TEST_SQLBACKEND", default="1") != "1",
                reason="SQL Backend skipped because $TEST_SQLBACKEND != 1 or sqlalchemy is not installed.",
            ),
            id="SQLBackend",
        ),
    )
if _CAN_IMPORT_IGRAPH:
    backend_test_params.append(
        pytest.param(
            (IGraphBackend, {}),
            marks=pytest.mark.skipif(
                os.environ.get("TEST_IGRAPHBACKEND", default="1") != "1",
                reason="IGraph Backend skipped because $TEST_IGRAPHBACKEND != 1 or igraph is not installed.",
            ),
            id="IGraphBackend",
        ),
    )
if _CAN_IMPORT_NETWORKIT:
    backend_test_params.append(
        pytest.param(
            (NetworkitBackend, {}),
            marks=pytest.mark.skipif(
                os.environ.get("TEST_NETWORKIT", default="1") != "1",
                reason="Networkit Backend skipped because $TEST_NETWORKIT != 1 or networkit is not installed.",
            ),
            id="NetworkitBackend",
        ),
    )

if os.environ.get("TEST_NETWORKITBACKEND") == "1":
    from ._networkit import NetworkitBackend

    backend_test_params.append(
        pytest.param(
            (NetworkitBackend, {}),
            marks=pytest.mark.skipif(
                os.environ.get("TEST_NETWORKITBACKEND") != "1",
                reason="Networkit Backend skipped because $TEST_NETWORKITBACKEND != 1.",
            ),
            id="NetworkitBackend",
        ),
    )

if os.environ.get("TEST_IGRAPHBACKEND") == "1":
    from ._igraph import IGraphBackend

    backend_test_params.append(
        pytest.param(
            (IGraphBackend, {}),
            marks=pytest.mark.skipif(
                os.environ.get("TEST_IGRAPHBACKEND") != "1",
                reason="Networkit Backend skipped because $TEST_IGRAPHBACKEND != 1.",
            ),
            id="IGraphBackend",
        ),
    )

# @pytest.mark.parametrize("backend", backend_test_params)
class TestBackendPersistence:
    def test_sqlite_persistence(self):
        if not _CAN_IMPORT_SQL:
            return

        dbpath = "grand_peristence_test_temp.db"
        url = "sqlite:///"+dbpath

        #arrange
        backend = SQLBackend(db_url=url, directed=True)
        node0 = backend.add_node("A",{"foo":"bar"})
        backend.commit()
        backend.close()
        #act
        backend = SQLBackend(db_url=url, directed=True)
        nodes = list(backend.all_nodes_as_iterable())
        #assert
        assert node0 in nodes
        #cleanup
        os.remove(dbpath)



@pytest.mark.parametrize("backend", backend_test_params)
class TestBackend:
    def test_can_create(self, backend):
        backend, kwargs = backend
        backend(**kwargs)

    def test_can_create_directed_and_undirected_backends(self, backend):
        backend, kwargs = backend
        b = backend(directed=True, **kwargs)
        assert b.is_directed() == True

        b = backend(directed=False, **kwargs)
        assert b.is_directed() == False

    def test_can_add_node(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        nxG = nx.Graph()
        G.nx.add_node("A", k="v")
        nxG.add_node("A", k="v")
        assert len(G.nx.nodes()) == len(nxG.nodes())
        G.nx.add_node("B", k="v")
        nxG.add_node("B", k="v")
        assert len(G.nx.nodes()) == len(nxG.nodes())

    def test_can_update_node(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        G.nx.add_node("A", k="v", z=3)
        G.nx.add_node("A", k="v2", x=4)
        assert G.nx.nodes["A"]["k"] == "v2"
        assert G.nx.nodes["A"]["x"] == 4
        assert G.nx.nodes["A"]["z"] == 3

    def test_can_add_edge(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        nxG = nx.Graph()
        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        assert len(G.nx.edges()) == len(nxG.edges())
        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        assert len(G.nx.edges()) == len(nxG.edges())

    def test_can_update_edge(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        G.nx.add_edge("A", "B", k="v", z=3)
        G.nx.add_edge("A", "B", k="v2", x=4)
        assert G.nx.get_edge_data("A", "B")["k"] == "v2"
        assert G.nx.get_edge_data("A", "B")["x"] == 4
        assert G.nx.get_edge_data("A", "B")["z"] == 3
        assert len(G.nx.nodes()) == 2

    def test_can_get_node(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        nxG = nx.Graph()
        md = dict(k="B")
        G.nx.add_node("A", **md)
        nxG.add_node("A", **md)
        assert G.nx.nodes["A"] == nxG.nodes["A"]

    def test_can_get_edge(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        nxG = nx.Graph()
        md = {"k": "B"}
        G.nx.add_edge("A", "B", **md)
        nxG.add_edge("A", "B", **md)
        assert G.nx.get_edge_data("A", "B") == nxG.get_edge_data("A", "B")

    def test_can_get_neighbors(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        nxG = nx.Graph()
        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        assert sorted([i for i in G.nx.neighbors("A")]) == sorted(
            [i for i in nxG.neighbors("A")]
        )
        assert sorted([i for i in G.nx.neighbors("B")]) == sorted(
            [i for i in nxG.neighbors("B")]
        )
        G.nx.add_edge("A", "C")
        nxG.add_edge("A", "C")
        assert sorted([i for i in G.nx.neighbors("A")]) == sorted(
            [i for i in nxG.neighbors("A")]
        )
        assert sorted([i for i in G.nx.neighbors("B")]) == sorted(
            [i for i in nxG.neighbors("B")]
        )
        assert sorted([i for i in G.nx.neighbors("C")]) == sorted(
            [i for i in nxG.neighbors("C")]
        )

    def test_undirected_adj(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        nxG = nx.Graph()
        assert G.nx._adj == nxG._adj
        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        assert G.nx._adj == nxG._adj

    def test_directed_adj(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(directed=True, **kwargs))
        nxG = nx.DiGraph()
        assert G.nx._adj == nxG._adj
        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        assert G.nx._adj == nxG._adj

    def test_can_traverse_undirected_graph(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        nxG = nx.Graph()
        md = dict(k="B")
        G.nx.add_edge("A", "B", **md)
        nxG.add_edge("A", "B", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        G.nx.add_edge("B", "C", **md)
        nxG.add_edge("B", "C", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        G.nx.add_edge("B", "D", **md)
        nxG.add_edge("B", "D", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        assert dict(nx.bfs_successors(G.nx, "C")) == dict(nx.bfs_successors(nxG, "C"))

    def test_can_traverse_directed_graph(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(directed=True, **kwargs))
        nxG = nx.DiGraph()
        md = dict(k="B")
        G.nx.add_edge("A", "B", **md)
        nxG.add_edge("A", "B", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        G.nx.add_edge("B", "C", **md)
        nxG.add_edge("B", "C", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        G.nx.add_edge("B", "D", **md)
        nxG.add_edge("B", "D", **md)
        assert dict(nx.bfs_successors(G.nx, "A")) == dict(nx.bfs_successors(nxG, "A"))
        assert dict(nx.bfs_successors(G.nx, "C")) == dict(nx.bfs_successors(nxG, "C"))

    def test_subgraph_isomorphism_undirected(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(directed=False, **kwargs))
        nxG = nx.Graph()

        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        G.nx.add_edge("B", "C")
        nxG.add_edge("B", "C")
        G.nx.add_edge("C", "A")
        nxG.add_edge("C", "A")

        from networkx.algorithms.isomorphism import GraphMatcher

        assert len(
            [i for i in GraphMatcher(G.nx, G.nx).subgraph_monomorphisms_iter()]
        ) == len([i for i in GraphMatcher(nxG, nxG).subgraph_monomorphisms_iter()])

    def test_subgraph_isomorphism_directed(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(directed=True, **kwargs))
        nxG = nx.DiGraph()

        G.nx.add_edge("A", "B")
        nxG.add_edge("A", "B")
        G.nx.add_edge("B", "C")
        nxG.add_edge("B", "C")
        G.nx.add_edge("C", "A")
        nxG.add_edge("C", "A")

        from networkx.algorithms.isomorphism import DiGraphMatcher

        assert len(
            [i for i in DiGraphMatcher(G.nx, G.nx).subgraph_monomorphisms_iter()]
        ) == len([i for i in DiGraphMatcher(nxG, nxG).subgraph_monomorphisms_iter()])

    def test_can_get_edge_metadata(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        G.nx.add_edge("foo", "bar", baz=True)
        assert list(G.nx.edges(data=True)) == [("foo", "bar", {"baz": True})]

    def test_edge_dne_raises(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        G.nx.add_edge("foo", "bar", baz=True)

        assert G.nx.has_edge("foo", "crab") == False
        assert G.nx.has_edge("foo", "bar") == True
        # assert G.nx.edges[("foo", "bar")] != None
        # with pytest.raises(Exception):
        #     G.nx.edges[("foo", "crab")]

    def test_reverse_edges_in_undirected(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(directed=False, **kwargs))
        G.nx.add_edge("foo", "bar", baz=True)

        assert G.nx.has_edge("foo", "bar") == True
        assert G.nx.has_edge("bar", "foo") == True

    def test_undirected_degree(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(directed=False, **kwargs))
        G.nx.add_edge("foo", "bar", baz=True)
        assert G.nx.degree("foo") == 1
        assert G.nx.degree("bar") == 1

    def test_directed_degree(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(directed=True, **kwargs))
        G.nx.add_edge("foo", "bar", baz=True)
        assert G.nx.degree("foo") == 1
        assert G.nx.degree("bar") == 0

    def test_undirected_degree_multiple(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(directed=False, **kwargs))
        G.nx.add_edge("foo", "bar", baz=True)
        G.nx.add_edge("foo", "baz", baz=True)
        assert G.nx.degree("foo") == 2
        assert G.nx.degree("bar") == 1
        assert G.nx.degree("baz") == 1

    def test_directed_degree_multiple(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(directed=True, **kwargs))
        G.nx.add_edge("foo", "bar", baz=True)
        G.nx.add_edge("foo", "baz", baz=True)
        assert G.nx.degree("foo") == 2
        assert G.nx.degree("bar") == 0
        assert G.nx.degree("baz") == 0
        assert G.nx.out_degree("foo") == 2
        assert G.nx.out_degree("bar") == 0
        assert G.nx.out_degree("baz") == 0
        assert G.nx.in_degree("foo") == 0
        assert G.nx.in_degree("bar") == 1
        assert G.nx.in_degree("baz") == 1

    def test_node_count(self, backend):
        backend, kwargs = backend
        G = Graph(backend=backend(**kwargs))
        G.nx.add_node("foo", bar=True)
        G.nx.add_node("bar", foo=True)
        assert len(G.nx) == 2


@pytest.mark.benchmark
@pytest.mark.parametrize("backend", backend_test_params)
def test_node_addition_performance(backend):
    backend, kwargs = backend
    G = Graph(backend=backend(directed=True, **kwargs))
    for i in range(1000):
        G.nx.add_node(i)
    assert len(G.nx) == 1000


@pytest.mark.benchmark
@pytest.mark.parametrize("backend", backend_test_params)
def test_get_density_performance(backend):
    backend, kwargs = backend
    G = Graph(backend=backend(directed=True, **kwargs))
    for i in range(1000):
        G.nx.add_node(i)
    for i in range(1000 - 1):
        G.nx.add_edge(i, i + 1)
    assert nx.density(G.nx) <= 0.005



"""
Grand graphs package.

"""

from typing import Optional
from .backends import Backend, NetworkXBackend
from .dialects import NetworkXDialect, IGraphDialect, NetworkitDialect


_DEFAULT_BACKEND = NetworkXBackend

__version__ = "0.5.1"


class Graph:
    """
    A grand.Graph enables you to manipulate a graph using multiple dialects.

    """

    nx: NetworkXDialect
    networkit: NetworkitDialect
    igraph: IGraphDialect

    def __init__(self, backend: Optional[Backend] = None, **backend_kwargs: dict):
        """
        Create a new grand.Graph.

        The only positional argument is the backend to use. All other arguments
        are passed to the backend's constructor, if a type is provided.
        Otherwise, kwargs are ignored.

        Arguments:
            backend (Backend): The backend to use. If none is provided, will
                default to _DEFAULT_BACKEND.

        """
        self.backend = backend or _DEFAULT_BACKEND

        # If you passed a class instead of an instance, instantiate it with
        # kwargs from the constructor:
        if isinstance(self.backend, type):
            self.backend = self.backend(**backend_kwargs)

        # Attach dialects:
        self.attach_dialect("nx", NetworkXDialect)
        self.attach_dialect("igraph", IGraphDialect)
        self.attach_dialect("networkit", NetworkitDialect)

    def attach_dialect(self, name: str, dialect: type):
        """
        Attach a dialect to the graph.

        Arguments:
            name (str): The name of the dialect.
            dialect (type): The dialect class to attach.

        """
        setattr(self, name, dialect(self))


class DiGraph(Graph):
    """
    A grand.DiGraph enables you to manipulate a directed graph. This is a
    convenience class that inherits from grand.Graph.

    """

    def __init__(self, backend: Optional[Backend] = None, **backend_kwargs: dict):
        """
        Create a new grand.DiGraph.

        The only positional argument is the backend to use. All other arguments
        are passed to the backend's constructor, if a type is provided.
        Otherwise, kwargs are ignored.

        Arguments:
            backend (Backend): The backend to use. If none is provided, will
                default to _DEFAULT_BACKEND.

        """
        super().__init__(backend, **{**backend_kwargs, "directed": True})



import io
import unittest

from .. import Graph
from . import (
    NetworkXDialect,
    IGraphDialect,
    NetworkitDialect,
    _GrandAdjacencyView,
    _GrandNodeAtlasView,
)
import networkx as nx


class TestIGraphDialect(unittest.TestCase):
    def test_igraph_verts(self):
        G = Graph()
        self.assertEqual(G.igraph.vs, [])
        G.nx.add_node("1")
        self.assertEqual(G.igraph.vs, [("1", {})])

    def test_add_verts(self):
        G = Graph()
        G.igraph.add_vertices(1)
        self.assertEqual(G.igraph.vs, [(0, {})])
        G.igraph.add_vertices(1)
        self.assertEqual(G.igraph.vs, [(0, {}), (1, {})])
        G.igraph.add_vertices(10)
        self.assertEqual(len(G.igraph.vs), 12)

    def test_igraph_edges(self):
        G = Graph()
        G.igraph.add_vertices(2)
        G.igraph.add_edges([(0, 1)])
        self.assertEqual(G.igraph.es, [(0, 1, {})])


class TestNetworkXHelpers(unittest.TestCase):
    def test_nx_adj_length(self):
        G = Graph()
        assert len(G.nx.adj) == 0
        G.nx.add_node("1")
        G.nx.add_edge("1", "2")
        assert len(G.nx.adj["1"]) == 1


class TestNetworkXDialect(unittest.TestCase):
    def test_nx_pred(self):
        G = Graph(directed=True)
        G.nx.add_edge("1", "2")
        G.nx.add_edge("1", "3")
        H = nx.DiGraph()
        H.add_edge("1", "2")
        H.add_edge("1", "3")
        self.assertEqual(G.nx.pred, H.pred)

    def test_nx_directed(self):
        G = Graph(directed=True)
        self.assertTrue(G.nx.is_directed())

        G = Graph(directed=False)
        self.assertFalse(G.nx.is_directed())

    def test_in_degree(self):
        G = Graph(directed=True)
        G.nx.add_edge("1", "2")
        G.nx.add_edge("1", "3")
        H = nx.DiGraph()
        H.add_edge("1", "2")
        H.add_edge("1", "3")
        self.assertEqual(dict(G.nx.in_degree()), dict(H.in_degree()))

    def test_out_degree(self):
        G = Graph(directed=True)
        G.nx.add_edge("1", "2")
        G.nx.add_edge("1", "3")
        H = nx.DiGraph()
        H.add_edge("1", "2")
        H.add_edge("1", "3")
        self.assertEqual(dict(G.nx.out_degree()), dict(H.out_degree()))

    def test_nx_edges(self):
        G = Graph(directed=True).nx
        H = nx.DiGraph()
        G.add_edge("1", "2")
        G.add_edge("1", "3")
        H.add_edge("1", "2")
        H.add_edge("1", "3")
        self.assertEqual(dict(G.edges), dict(H.edges))
        self.assertEqual(dict(G.edges()), dict(H.edges()))
        self.assertEqual(list(G.edges["1", "2"]), list(H.edges["1", "2"]))

    def test_nx_export(self):
        gg = Graph()
        f = io.BytesIO()
        nx.write_graphml(gg.nx, f)


class TestNetworkitDialect(unittest.TestCase):
    def test_add_verts(self):
        G = Graph()
        u = G.networkit.addNode()
        assert G.networkit.hasNode(u)
        assert not G.networkit.hasNode("X")

    def test_add_edges(self):
        G = Graph()
        u = G.networkit.addNode()
        v = G.networkit.addNode()
        assert len(G.networkit.edges()) == 0
        assert G.networkit.numberOfEdges() == 0
        G.networkit.addEdge(u, v)
        assert G.networkit.hasEdge(u, v)
        assert len(G.networkit.edges()) == 1
        assert G.networkit.numberOfEdges() == 1
        assert G.networkit.numberOfNodes() == 2

    def test_nodes(self):
        G = Graph()
        G.networkit.addNode()
        G.networkit.addNode()
        self.assertEqual(len(G.networkit.nodes()), 2)
        assert G.networkit.numberOfNodes() == 2

    def test_undirected_degree(self):
        G = Graph(directed=False)
        assert G.networkit.numberOfNodes() == 0
        assert G.networkit.numberOfEdges() == 0
        u = G.networkit.addNode()
        v = G.networkit.addNode()
        assert G.networkit.numberOfNodes() == 2
        assert G.networkit.numberOfEdges() == 0
        assert G.networkit.degree(u) == 0
        assert G.networkit.degree(v) == 0
        G.networkit.addEdge(u, v)
        assert G.networkit.degree(u) == 1
        assert G.networkit.degree(v) == 1
        assert G.networkit.numberOfEdges() == 1

    def test_directed_degree(self):
        G = Graph(directed=True)
        assert G.networkit.numberOfNodes() == 0
        assert G.networkit.numberOfEdges() == 0
        u = G.networkit.addNode()
        v = G.networkit.addNode()
        assert G.networkit.numberOfNodes() == 2
        assert G.networkit.numberOfEdges() == 0
        assert G.networkit.degreeIn(u) == 0
        assert G.networkit.degreeOut(u) == 0
        assert G.networkit.degreeIn(v) == 0
        assert G.networkit.degreeOut(v) == 0
        G.networkit.addEdge(u, v)
        assert G.networkit.degreeIn(u) == 0
        assert G.networkit.degreeOut(u) == 1
        assert G.networkit.degreeIn(v) == 1
        assert G.networkit.degreeOut(v) == 0
        assert G.networkit.numberOfEdges() == 1

    def test_undirected_density(self):
        G = Graph(directed=False)
        u = G.networkit.addNode()
        v = G.networkit.addNode()
        G.networkit.addEdge(u, v)
        assert G.networkit.density() == 1

    def test_directed_density(self):
        G = Graph(directed=True)
        u = G.networkit.addNode()
        v = G.networkit.addNode()
        G.networkit.addEdge(u, v)
        assert G.networkit.density() == 0.5
        G.networkit.addEdge(v, u)
        assert G.networkit.density() == 1



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="grand-graph",
    version="0.5.1",
    author="Jordan Matelsky",
    author_email="opensource@matelsky.com",
    description="Graph database wrapper for non-graph datastores",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aplbrain/grand",
    packages=setuptools.find_packages(),
    install_requires=["networkx>=2.4", "numpy", "pandas", "cachetools"],
    extras_require={
        "sql": ["SQLAlchemy>=1.3"],
        "dynamodb": ["boto3"],
        "igraph": ["igraph"],
        "networkit": ["cmake", "cython", "networkit"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from typing import Hashable, Generator
import time

import pandas as pd

from .backend import Backend


class DataFrameBackend(Backend):
    def __init__(
        self,
        directed: bool = False,
        edge_df: pd.DataFrame = None,
        node_df: pd.DataFrame = None,
        edge_df_source_column: str = "Source",
        edge_df_target_column: str = "Target",
        node_df_id_column: str = "id",
    ):
        """
        Create a new DataFrame backend.

        You must pass an edgelist. A nodelist is optional.

        Arguments:
            edge_df (pd.DataFrame): An edgelist dataframe with one edge per row
            directed (bool: False): Whether the graph is directed
            node_df (pd.DataFrame): A node metadata lookup
            edge_df_source_column (str): The name of the column in `edge_df` to
                use as the source of edges
            edge_df_target_column (str): The name of the column in `edge_df` to
                use as the target of edges
            node_df_id_column (str): The name of the column in `node_df` to
                use as the node ID
        """
        self._directed = directed
        self._edge_df = edge_df or pd.DataFrame(
            columns=[edge_df_source_column, edge_df_target_column]
        )
        self._node_df = node_df or None
        self._edge_df_source_column = edge_df_source_column
        self._edge_df_target_column = edge_df_target_column
        self._node_df_id_column = node_df_id_column

    def is_directed(self) -> bool:
        """
        Return True if the backend graph is directed.

        Arguments:
            None

        Returns:
            bool: True if the backend graph is directed.

        """
        return self._directed

    def teardown(self, yes_i_am_sure: bool = False):
        """
        Tear down this graph, deleting all evidence it once was here.

        """
        return

    def add_node(self, node_name: Hashable, metadata: dict) -> Hashable:
        """
        Add a new node to the graph.

        Insert a new document into the nodes table.

        Arguments:
            node_name (Hashable): The ID of the node
            metadata (dict: None): An optional dictionary of metadata

        Returns:
            Hashable: The ID of this node, as inserted

        """

        # Add a new row to the nodes table:
        if self._node_df is None:
            self._node_df = pd.DataFrame(
                [
                    {
                        self._node_df_id_column: node_name,
                        **metadata,
                    }
                ],
                columns=[
                    self._node_df_id_column,
                    *metadata.keys(),
                ],
            )
            self._node_df.set_index(self._node_df_id_column, inplace=True)
        else:
            if self.has_node(node_name):
                existing_metadata = self.get_node_by_id(node_name)
                existing_metadata.update(metadata)
                for k, v in existing_metadata.items():
                    self._node_df.at[node_name, k] = v
            else:
                # Insert a new row:
                self._node_df = pd.concat(
                    [self._node_df, pd.DataFrame([{node_name: metadata}]).T]
                )

        return node_name

    def all_nodes_as_iterable(self, include_metadata: bool = False):
        """
        Get a generator of all of the nodes in this graph.

        Arguments:
            include_metadata (bool: False): Whether to include node metadata in
                the response

        Returns:
            Generator: A generator of all nodes (arbitrary sort)

        """
        if self._node_df is not None:
            return [
                (
                    (
                        node_id,
                        row.to_dict(),
                    )
                    if include_metadata
                    else node_id
                )
                for node_id, row in self._node_df.iterrows()
            ]

        else:
            return [
                (node_id, {}) if include_metadata else node_id
                for node_id in self._edge_df[self._edge_df_source_column]
            ] + [
                (node_id, {}) if include_metadata else node_id
                for node_id in self._edge_df[self._edge_df_target_column]
            ]

    def has_node(self, u: Hashable) -> bool:
        """
        Return true if the node exists in the graph.

        Arguments:
            u (Hashable): The ID of the node to check

        Returns:
            bool: True if the node exists
        """
        if self._node_df is not None:
            return u in self._node_df.index

        return u in (self._edge_df[self._edge_df_source_column]) or u in (
            self._edge_df[self._edge_df_target_column]
        )

    def add_edge(self, u: Hashable, v: Hashable, metadata: dict):
        """
        Add a new edge to the graph between two nodes.

        If the graph is directed, this edge will start (source) at the `u` node
        and end (target) at the `v` node.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID
            metadata (dict): Optional metadata to associate with the edge

        Returns:
            Hashable: The edge ID, as inserted.

        """

        if not self.has_node(u):
            self.add_node(u, {})
        if not self.has_node(v):
            self.add_node(v, {})

        if self._has_edge(u, v):
            # Update the existing edge:
            for k, m in metadata.items():
                if self._directed:
                    self._edge_df.loc[
                        (self._edge_df[self._edge_df_source_column] == u)
                        & (self._edge_df[self._edge_df_target_column] == v),
                        k,
                    ] = m
                else:
                    # Check for the edge in both directions:
                    self._edge_df.loc[
                        (self._edge_df[self._edge_df_source_column] == u)
                        & (self._edge_df[self._edge_df_target_column] == v),
                        k,
                    ] = m
                    self._edge_df.loc[
                        (self._edge_df[self._edge_df_source_column] == v)
                        & (self._edge_df[self._edge_df_target_column] == u),
                        k,
                    ] = m
        else:
            row = {
                self._edge_df_source_column: u,
                self._edge_df_target_column: v,
                **metadata,
            }
            self._edge_df.loc[len(self._edge_df)] = None
            for k, m in row.items():
                self._edge_df.loc[len(self._edge_df) - 1, k] = m
        return (u, v)

    def _has_edge(self, u: Hashable, v: Hashable) -> bool:
        """
        Return true if the edge exists in the graph.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID

        Returns:
            bool: True if the edge exists
        """
        if self._directed:
            return (
                (self._edge_df[self._edge_df_source_column] == u)
                & (self._edge_df[self._edge_df_target_column] == v)
            ).any()
        else:
            return (
                (self._edge_df[self._edge_df_source_column] == u)
                & (self._edge_df[self._edge_df_target_column] == v)
            ).any() or (
                (self._edge_df[self._edge_df_source_column] == v)
                & (self._edge_df[self._edge_df_target_column] == u)
            ).any()

    def all_edges_as_iterable(self, include_metadata: bool = False) -> Generator:
        """
        Get a list of all edges in this graph, arbitrary sort.

        Arguments:
            include_metadata (bool: False): Whether to include edge metadata

        Returns:
            Generator: A generator of all edges (arbitrary sort)

        """
        for _, row in self._edge_df.iterrows():
            if include_metadata:
                yield (
                    row[self._edge_df_source_column],
                    row[self._edge_df_target_column],
                    dict(row),
                )
            else:
                yield (
                    row[self._edge_df_source_column],
                    row[self._edge_df_target_column],
                )

    def get_node_by_id(self, node_name: Hashable):
        """
        Return the data associated with a node.

        Arguments:
            node_name (Hashable): The node ID to look up

        Returns:
            dict: The metadata associated with this node

        """
        if self._node_df is not None:
            res = (self._node_df.loc[node_name]).to_dict()
            return res.get(0, res)

        return {}

    def get_edge_by_id(self, u: Hashable, v: Hashable):
        """
        Get an edge by its source and target IDs.

        Arguments:
            u (Hashable): The source node ID
            v (Hashable): The target node ID

        Returns:
            dict: Metadata associated with this edge

        """
        if self._directed:
            result = self._edge_df[
                (self._edge_df[self._edge_df_source_column] == u)
                & (self._edge_df[self._edge_df_target_column] == v)
            ]
            if len(result):
                return self._edge_as_dict(result.iloc[0])

        else:
            left = self._edge_df[
                (self._edge_df[self._edge_df_source_column] == u)
                & (self._edge_df[self._edge_df_target_column] == v)
            ]
            if len(left):
                return self._edge_as_dict(left.iloc[0])
            right = self._edge_df[
                (self._edge_df[self._edge_df_source_column] == v)
                & (self._edge_df[self._edge_df_target_column] == u)
            ]
            if len(right):
                return self._edge_as_dict(right.iloc[0])

    def get_node_neighbors(self, u: Hashable, include_metadata: bool = False):
        """
        Get a generator of all downstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """

        if include_metadata:
            if self._directed:
                return {
                    (r[self._edge_df_target_column]): self._edge_as_dict(r)
                    for _, r in self._edge_df[
                        (self._edge_df[self._edge_df_source_column] == u)
                    ].iterrows()
                }
            else:
                return {
                    (
                        r[self._edge_df_source_column]
                        if r[self._edge_df_source_column] != u
                        else r[self._edge_df_target_column]
                    ): self._edge_as_dict(r)
                    for _, r in self._edge_df[
                        (self._edge_df[self._edge_df_source_column] == u)
                        | (self._edge_df[self._edge_df_target_column] == u)
                    ].iterrows()
                }

        if self._directed:
            return iter(
                [
                    row[self._edge_df_target_column]
                    for _, row in self._edge_df[
                        (self._edge_df[self._edge_df_source_column] == u)
                    ].iterrows()
                ]
            )
        else:
            return iter(
                [
                    (
                        row[self._edge_df_source_column]
                        if row[self._edge_df_source_column] != u
                        else row[self._edge_df_target_column]
                    )
                    for _, row in self._edge_df[
                        (self._edge_df[self._edge_df_source_column] == u)
                        | (self._edge_df[self._edge_df_target_column] == u)
                    ].iterrows()
                ]
            )

    def _edge_as_dict(self, row):
        """
        Convert an edge row to a dictionary.

        Arguments:
            row (pandas.Series): The edge row

        Returns:
            dict: The edge metadata

        """
        r = row.to_dict()
        r.pop(self._edge_df_source_column)
        r.pop(self._edge_df_target_column)
        return r

    def get_node_predecessors(self, u: Hashable, include_metadata: bool = False):
        """
        Get a generator of all upstream nodes from this node.

        Arguments:
            u (Hashable): The source node ID

        Returns:
            Generator

        """

        if include_metadata:
            if self._directed:
                return {
                    (
                        r[self._edge_df_target_column]
                        if r[self._edge_df_target_column] != u
                        else r[self._edge_df_source_column]
                    ): self._edge_as_dict(r)
                    for _, r in self._edge_df[
                        (self._edge_df[self._edge_df_target_column] == u)
                    ].iterrows()
                }
            else:
                return {
                    (
                        r[self._edge_df_target_column]
                        if r[self._edge_df_target_column] != u
                        else r[self._edge_df_source_column]
                    ): self._edge_as_dict(r)
                    for _, r in self._edge_df[
                        (self._edge_df[self._edge_df_target_column] == u)
                        | (self._edge_df[self._edge_df_source_column] == u)
                    ].iterrows()
                }

        if self._directed:
            return iter(
                [
                    row[self._edge_df_source_column]
                    for _, row in self._edge_df[
                        (self._edge_df[self._edge_df_target_column] == u)
                    ].iterrows()
                ]
            )
        else:
            return iter(
                [
                    (
                        row[self._edge_df_source_column]
                        if row[self._edge_df_target_column] != u
                        else row[self._edge_df_target_column]
                    )
                    for _, row in self._edge_df[
                        (self._edge_df[self._edge_df_target_column] == u)
                        | (self._edge_df[self._edge_df_source_column] == u)
                    ].iterrows()
                ]
            )

    def get_node_count(self) -> int:
        """
        Get an integer count of the number of nodes in this graph.

        Arguments:
            None

        Returns:
            int: The count of nodes

        """
        if self._node_df is not None:
            return len(self._node_df)
        # Return number of unique sources intersected with number of unique targets
        return len(
            set(self._edge_df[self._edge_df_source_column]).intersection(
                set(self._edge_df[self._edge_df_target_column])
            )
        )

    def get_edge_count(self) -> int:
        """
        Get an integer count of the number of edges in this graph.

        Arguments:
            None

        Returns:
            int: The count of edges

        """
        return len(self._edge_df)

    def ingest_from_edgelist_dataframe(
        self, edgelist: pd.DataFrame, source_column: str, target_column: str
    ) -> dict:
        """
        Ingest an edgelist from a Pandas DataFrame.

        """
        # Produce edge list:

        edge_tic = time.time()
        self._edge_df = edgelist
        self._edge_df_source_column = source_column
        self._edge_df_target_column = target_column

        return {
            "edge_duration": time.time() - edge_tic,
        }
